# Copyright (c) 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""PIPS version info.
"""

def pipsver(*args):
    """ Returns PIPS version info for current installation.

    @author: Ray Zimmerman (PSERC Cornell)
    """
    ver = {'Name': 'PIPS',
           'Version': '1.0',
           'Release':  '',
           'Date': '07-Feb-2011'}

    return ver
