# Copyright (c) 1996-2015 PSERC. All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

"""PYPOWER version info.
"""

def ppver(*args):
    """ Returns PYPOWER version info for current installation.

    @author: Ray Zimmerman (PSERC Cornell)
    """

    ver = {'Name': 'PYPOWER',
           'Version': '5.1.17',
           'Release':  '',
           'Date': '02-Sept-2024'}

    return ver
